
#include <iostream>
#include <vector>
#include <string>

//#include "GpuDemoUtils.hpp"

#include "chrono/core/ChGlobal.h" // Creates workspace and application
#include "chrono/physics/ChSystemSMC.h" // Set's basic physical system rules, i.e. contact force, adhesion force, Calls Ch.Systemh -> defines all params and phsyics of workspace
#include "chrono/physics/ChBody.h" // Defines rigidy body physics, references a lot of files surrounding this
#include "chrono/physics/ChForce.h" // How forces are applied to rigid bodies
#include "chrono/timestepper/ChTimestepper.h" // Base class for timesteppers, i.e., time integrators that can advance a system state.
#include "chrono/utils/ChUtilsSamplers.h" // Samples data based on a set domain definitions
#include "chrono/utils/ChUtilsCreators.h" // Utitility functions to facilitate adding contact and visualization geometry to a body. Adds all the geometries.
#include "chrono/assets/ChSphereShape.h" // Class for the visualization of the sprhere
#include "chrono/assets/ChConeShape.h" // Call for the visualization of the cone
#include "chrono_gpu/ChGpuData.h" // Setting paths to the gpu filespace
#include "chrono_gpu/physics/ChSystemGpu.h" // Constructs physical system for all of the granular media.
#include "chrono_gpu/utils/ChGpuJsonParser.h" // Simulation parameters
#include "chrono_gpu/utils/ChGpuVisualization.h" // Just visualizing the simulation
#include "chrono_thirdparty/filesystem/path.h" // C++ work around, purpose unclear
#include "chrono/assets/ChColorAsset.h"
#include "chrono/assets/ChCamera.h"
#include "chrono/physics/ChMaterialSurfaceSMC.h"
#include "chrono/physics/ChParticlesClones.h"
#include "chrono_gpu/physics/ChSystemGpu.h"
#include "chrono_gpu/physics/ChSystemGpu_impl.h"
#include "chrono_gpu/physics/ChSystemGpuMesh_impl.h"
#include "chrono_gpu/utils/ChGpuUtilities.h"


#include "chrono_gpu/physics/ChSystemGpu.h"
#include "chrono_gpu/physics/ChSystemGpu_impl.h"
#include "chrono_gpu/physics/ChSystemGpuMesh_impl.h"

#include "chrono_gpu/utils/ChGpuUtilities.h"

using namespace chrono;
using namespace chrono::gpu;
//using namespace chrono::postprocess;


// If below vaues can't be read, copy-paste them into main function

// Output frequency
float out_fps = 50;

// Enable/disable run-time visualization (if Chrono::OpenGL is available)
bool render = true;
float render_fps = 2000;


// Defines position of cone and tracks the position and velocity of the cone as it moves.
 void writeMeshFrames(std::ostringstream& outstream, ChBody& body, std::string obj_name, float mesh_scaling) {
     outstream << obj_name << ",";

     // Get frame position
     ChFrame<> body_frame = body.GetFrame_REF_to_abs();
     ChQuaternion<> rot = body_frame.GetRot();
     ChVector<> pos = body_frame.GetPos();

     // Get basis vectors
     ChVector<> vx = rot.GetXaxis();
     ChVector<> vy = rot.GetYaxis();
     ChVector<> vz = rot.GetZaxis();

     // Output in order
     outstream << pos.x() << ",";
     outstream << pos.y() << ",";
     outstream << pos.z() << ",";
     outstream << vx.x() << ",";
     outstream << vx.y() << ",";
     outstream << vx.z() << ",";
     outstream << vy.x() << ",";
     outstream << vy.y() << ",";
     outstream << vy.z() << ",";
     outstream << vz.x() << ",";
     outstream << vz.y() << ",";
     outstream << vz.z() << ",";
     outstream << mesh_scaling << "," << mesh_scaling << "," << mesh_scaling;
     outstream << "\n";
 }

int main(int argc, char* argv[]) {


    ChGpuSimulationParameters params; // Setting parameters
     if (argc != 2 || ParseJSON(gpu::GetDataFile(argv[1]), params) == false) {
         std::cout << "Usage:\n./cone_drop_gpu <json_file>" << std::endl; // Pulling Json files where parameters are stored
         return 1;
     }

    float iteration_step = params.step_size; // creating time-step
    float cone_radius = .025; // radius of cone (cm)
    float cone_height = .050; // height of cone (cm)

    ChSystemGpuMesh gpu_sys(params.sphere_radius, params.sphere_density, // calling gpu_sys function and implementing values from the json file
                            ChVector<>(params.box_X, params.box_Y, params.box_Z));

    double fill_bottom = -params.box_Z / 2.0;
    double fill_top = params.box_Z / 4.0; // defining variables based on json file

    chrono::utils::PDSampler<float> sampler(2.4f * params.sphere_radius);
    // chrono::utils::HCPSampler<float> sampler(2.05 * sphere_radius);

    // leave a 4cm margin at edges of sampling
    ChVector<> hdims(params.box_X / 2 - 0.002, params.box_Y / 2 - 0.002, 0); //Changed from 4 to .04
    ChVector<> center(0, 0, fill_bottom + 2.0 * params.sphere_radius);		// creating box geometry and filling it with particles
    std::vector<ChVector<float>> body_points;

    // Shift up for bottom of box
    center.z() += 3 * params.sphere_radius;
    while (center.z() < fill_top) {
        std::cout << "Create layer at " << center.z() << std::endl;
        auto points = sampler.SampleBox(center, hdims);
        body_points.insert(body_points.end(), points.begin(), points.end());
        center.z() += 2.05 * params.sphere_radius;
    }

    gpu_sys.SetParticles(body_points);
    gpu_sys.SetBDFixed(true);

    /*Send sinusoidal wave through particle bed in x direction after 0.5 seconds?
    std::function<double3(float)> pos_func_wave = [&params](float t ) {
        double3 pos = {0, 0, 0};

        double t0 = 0.5;
        double freq = CH_C_PI / 4;

        if (t > t0) {
            pos.x = 0.1 * params.box_X * std::sin((t - t0) * freq);
        }
        return pos;
    };*/

    //gpu_sys.setBDWallsMotionFunction(pos_func_wave);


    // Defining the granular media
    gpu_sys.SetKn_SPH2SPH(params.normalStiffS2S);
    gpu_sys.SetKn_SPH2WALL(params.normalStiffS2W);
    gpu_sys.SetKn_SPH2MESH(params.normalStiffS2M);

    gpu_sys.SetGn_SPH2SPH(params.normalDampS2S);
    gpu_sys.SetGn_SPH2WALL(params.normalDampS2W);
    gpu_sys.SetGn_SPH2MESH(params.normalDampS2M);

    gpu_sys.SetKt_SPH2SPH(params.tangentStiffS2S);
    gpu_sys.SetKt_SPH2WALL(params.tangentStiffS2W);
    gpu_sys.SetKt_SPH2MESH(params.tangentStiffS2M);

    gpu_sys.SetGt_SPH2SPH(params.tangentDampS2S);
    gpu_sys.SetGt_SPH2WALL(params.tangentDampS2W);
    gpu_sys.SetGt_SPH2MESH(params.tangentDampS2M);

    gpu_sys.SetCohesionRatio(params.cohesion_ratio);
    gpu_sys.SetAdhesionRatio_SPH2MESH(params.adhesion_ratio_s2m);
    gpu_sys.SetAdhesionRatio_SPH2WALL(params.adhesion_ratio_s2w);

    gpu_sys.SetGravitationalAcceleration(ChVector<float>(params.grav_X, params.grav_Y, params.grav_Z));

    gpu_sys.SetFixedStepSize(params.step_size);
    gpu_sys.SetFrictionMode(CHGPU_FRICTION_MODE::MULTI_STEP);
    gpu_sys.SetTimeIntegrator(CHGPU_TIME_INTEGRATOR::CENTERED_DIFFERENCE);

    gpu_sys.SetStaticFrictionCoeff_SPH2SPH(params.static_friction_coeffS2S);
    gpu_sys.SetStaticFrictionCoeff_SPH2WALL(params.static_friction_coeffS2W);
    gpu_sys.SetStaticFrictionCoeff_SPH2MESH(params.static_friction_coeffS2M);

    // gpu_sys.SetRollingCoeff_SPH2SPH(params.rolling_friction_coeffS2S);
    // gpu_sys.SetRollingCoeff_SPH2WALL(params.rolling_friction_coeffS2W);
    // gpu_sys.SetRollingCoeff_SPH2MESH(params.rolling_friction_coeffS2M);

    std::string mesh_filename("../../cone_drop_1/Cone.obj");
    std::vector<string> mesh_filenames(1, mesh_filename);

    std::cout << "Setting mesh translation" << "\n";

    std::vector<float3> mesh_translations(1, make_float3(0.f, 0.f, 0.f));

    std::cout << "Setting mesh_rotscales" << "\n";

    std::vector<ChMatrix33<float>> mesh_rotscales(1, ChMatrix33<float>(cone_radius)); //Setting rotational scales of Mesh, might look funky in practice, will rotate direction here. 

    std::cout << "Calculating cone mass, adding to mesh mass" << "\n";
    float cone_density = cone_density;
    float cone_mass = (float)(4.f * CH_C_PI * cone_radius * cone_radius * cone_height * cone_density / 3.f); // Calculate baed on cone size.
    std::vector<float> mesh_masses(1, cone_mass);

    std::cout << "Loading Mesh" << "\n";
    gpu_sys.LoadMeshes(mesh_filenames, mesh_rotscales, mesh_translations, mesh_masses); // This function in ChSystemGpu.h

    gpu_sys.SetParticleOutputMode(params.write_mode);
    gpu_sys.SetVerbosity(params.verbose);
    std::cout << "verbose: " << static_cast<int>(params.verbose) << std::endl;
    gpu_sys.SetRecordingContactInfo(true);


    std::string out_dir = GetChronoOutputPath() + "GPU/";
    filesystem::create_directory(filesystem::path(out_dir));
    out_dir = out_dir + params.output_dir;
    filesystem::create_directory(filesystem::path(out_dir));

    std::cout << gpu_sys.GetNumMeshes() << " meshes" << std::endl;
    std::cout << "Initializing" << "\n";
    gpu_sys.Initialize();

    // Create rigid cone body simulation
    ChSystemSMC sys_cone;
    sys_cone.SetContactForceModel(ChSystemSMC::ContactForceModel::Hooke);
    sys_cone.SetTimestepperType(ChTimestepper::Type::EULER_IMPLICIT);
    sys_cone.Set_G_acc(ChVector<>(0, 0, -9.80));

    double inertia = 3.0 / 10.0 * cone_mass * cone_radius * cone_radius;
    ChVector<> cone_initial_pos(0, 0, fill_top + cone_radius + 2 * params.sphere_radius);

    std::shared_ptr<ChBody> cone_body(sys_cone.NewBody());
    cone_body->SetMass(cone_mass);
    cone_body->SetInertiaXX(ChVector<>(inertia, inertia, inertia));
    cone_body->SetPos(cone_initial_pos);
    auto con = chrono_types::make_shared<ChConeShape>();
    con->GetConeGeometry().rad = cone_radius; //cm radius of cone
    con->GetConeGeometry().center = ChVector<>(0, 0, 0); //fill_top + cone_radius + 2 * params.sphere_radius);
    cone_body->AddAsset(con);

    sys_cone.AddBody(cone_body);

    std::cout << "Output at " << out_fps << " FPS" << std::endl;
    std::cout << "Rendering at " << render_fps << " FPS" << std::endl;
    unsigned int out_steps = (unsigned int)(1.0 / (out_fps * iteration_step));
    unsigned int render_steps = (unsigned int)(1.0 / (render_fps * iteration_step));
    unsigned int total_frames = (unsigned int)(params.time_end * out_fps);

    int currframe = 0;
    unsigned int curr_step = 0;

    std::cout << "Running Simulation" << "\n";
    // Collecting data from previously defined parameters
    clock_t start = std::clock();
    for (double t = 0; t < (double)params.time_end; t += iteration_step, curr_step++) {
        gpu_sys.ApplyMeshMotion(0, cone_body->GetPos(), cone_body->GetRot(), cone_body->GetPos_dt(),
                                cone_body->GetWvel_par());

        ChVector<> cone_force;
        ChVector<> cone_torque;
        gpu_sys.CollectMeshContactForces(0, cone_force, cone_torque);

        //force = ChVector<>(gpu_sys.force);


        cone_body->Empty_forces_accumulators();
        cone_body->Accumulate_force(cone_force, cone_body->GetPos(), false);
        cone_body->Accumulate_torque(cone_torque, false);

        if (curr_step % out_steps == 0) {
            std::cout << "Output frame " << currframe + 1 << " of " << total_frames << std::endl;
            char filename[100];
            sprintf(filename, "%s/step%06d", out_dir.c_str(), currframe);
            gpu_sys.WriteParticleFile(std::string(filename));

            //std::cout << "starting contact forces" << "\n";
            //trying to get contact forces
            char contactFilename[100];
            sprintf(contactFilename, "%s/contact%06d", out_dir.c_str(), currframe);
            gpu_sys.WriteContactInfoFile(std::string(contactFilename));
            //std::cout << "ending contact forces" << "\n";

            gpu_sys.WriteMeshes(std::string(filename));
            currframe++;

            // disable meshframes output, for it may be confusing for users dealing with Chrono::Gpu only
            

            // disable meshframes output, for it may be confusing for users dealing with Chrono::Gpu only
            // std::string mesh_output = std::string(filename) + "_meshframes.csv";
            // std::ofstream meshfile(mesh_output);
            // std::ostringstream outstream;
            // outstream << "mesh_name,dx,dy,dz,x1,x2,x3,y1,y2,y3,z1,z2,z3,sx,sy,sz\n";
            // writeMeshFrames(outstream, *cone_body, mesh_filename, cone_radius);
            // meshfile << outstream.str();
            
            
        }

        gpu_sys.AdvanceSimulation(iteration_step);
        sys_cone.DoStepDynamics(iteration_step);
    }

    clock_t end = std::clock();
    double total_time = ((double)(end - start)) / CLOCKS_PER_SEC;

    std::cout << "Time: " << total_time << " seconds" << std::endl;

    return 0;
}
